<?php

namespace Drupal\entity_reference_preview\Plugin\Field\FieldFormatter;

use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\RevisionableStorageInterface;
use Drupal\Core\Entity\TranslatableRevisionableStorageInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceEntityFormatter;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\entity_reference_preview\Events\PreviewNegotiationSubscriber;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the 'entity reference rendered entity' formatter.
 *
 * @FieldFormatter(
 *   id = "entity_reference_entity_view_preview",
 *   label = @Translation("Rendered entity (with preview)"),
 *   description = @Translation("Display the referenced entities rendered by
 *   entity_view() matching the page moderation state."), field_types = {
 *     "entity_reference"
 *   }
 * )
 */
final class EntityReferenceEntityPreviewFormatter extends EntityReferenceEntityFormatter {

  /**
   * Detects the previewing state.
   *
   * @var bool
   */
  private $isPreviewing;

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  private $languageManager;

  /**
   * Detects the access to the draft indicator.
   *
   * @var bool
   */
  private $canSeeIndicator;

  /**
   * EntityReferenceEntityPreviewFormatter constructor.
   *
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the formatter is associated.
   * @param array $settings
   *   The formatter settings.
   * @param string $label
   *   The formatter label display setting.
   * @param string $view_mode
   *   The view mode.
   * @param array $third_party_settings
   *   Any third party settings settings.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entity_display_repository
   *   The entity display repository.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager.
   * @param bool $can_see_indicator
   *   Boolean weather or not current user can see the indicator.
   * @param bool $is_previewing
   *   Boolean indicating weather or not the request is previewing.
   */
  public function __construct(
    $plugin_id,
    $plugin_definition,
    FieldDefinitionInterface $field_definition,
    array $settings,
    $label,
    $view_mode,
    array $third_party_settings,
    LoggerChannelFactoryInterface $logger_factory,
    EntityTypeManagerInterface $entity_type_manager,
    EntityDisplayRepositoryInterface $entity_display_repository,
    LanguageManagerInterface $language_manager,
    bool $can_see_indicator,
    bool $is_previewing
  ) {
    parent::__construct(
      $plugin_id,
      $plugin_definition,
      $field_definition,
      $settings,
      $label,
      $view_mode,
      $third_party_settings,
      $logger_factory,
      $entity_type_manager,
      $entity_display_repository
    );
    $this->languageManager = $language_manager;
    $this->canSeeIndicator = $can_see_indicator;
    $this->isPreviewing = $is_previewing;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $can_see_indicator = $container->get('current_user')
      ->hasPermission('view entity_reference_preview indicator');
    $indicator_enabled = $container->get('config.factory')
      ->get('entity_reference_preview.settings')
      ->get('enableDraftIndicator');
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('logger.factory'),
      $container->get('entity_type.manager'),
      $container->get('entity_display.repository'),
      $container->get('language_manager'),
      $can_see_indicator && $indicator_enabled,
      (bool) $container->get('request_stack')
        ->getCurrentRequest()
        ->get(PreviewNegotiationSubscriber::IS_PREVIEWING)
    );
  }

  /**
   * {@inheritdoc}
   */
  public function prepareView(array $entities_items) {
    $this->isPreviewing
      ? $this->preparePreviewView($entities_items)
      : parent::prepareView($entities_items);
  }

  /**
   * Prepares the field items with the entity property loading the revision.
   *
   * The appropriate revision is negotiated based on the latest revision, the
   * render language and its parents.
   *
   * @param array $entities_items
   *   The entity items containing the field items.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *
   * @see \Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceFormatterBase::prepareView()
   */
  private function preparePreviewView(array $entities_items) {
    // Collect entity IDs to load. For performance, we want to use a single
    // "multiple entity load" to load all the entities for the multiple
    // "entity reference item lists" being displayed. We thus cannot use
    // \Drupal\Core\Field\EntityReferenceFieldItemList::referencedEntities().
    $revision_ids = [];
    foreach ($entities_items as $items) {
      foreach ($items as $item) {
        // To avoid trying to reload non-existent entities in
        // getEntitiesToView(), explicitly mark the items where $item->entity
        // contains a valid entity ready for display. All items are initialized
        // at FALSE.
        $item->_loaded = FALSE;
        if ($this->needsEntityLoad($item)) {
          // Find the latest revisions.
          $revision_ids[$item->target_id] = $this->findRevisionId($item->target_id);
        }
      }
    }
    if ($revision_ids) {
      $target_type = $this->getFieldSetting('target_type');
      $entity_storage = $this->entityTypeManager->getStorage($target_type);
      $target_entities = $entity_storage->loadMultipleRevisions($revision_ids);
    }

    // For each item, pre-populate the loaded entity in $item->entity, and set
    // the 'loaded' flag.
    foreach ($entities_items as $items) {
      foreach ($items as $item) {
        $revision_id = $revision_ids[$item->target_id];
        if (isset($target_entities[$revision_id])) {
          $item->entity = $target_entities[$revision_id];
          $item->_loaded = TRUE;
        }
        elseif ($item->hasNewEntity()) {
          $item->_loaded = TRUE;
        }
      }
    }
  }

  /**
   * Calculates the latest revision ID for the languages on the render node.
   *
   * @param string|int $entity_id
   *   The entity ID.
   *
   * @return int|string|null
   *   The revision ID if any.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function findRevisionId($entity_id) {
    $language_manager = \Drupal::languageManager();
    $langcode = $language_manager->getCurrentLanguage()->getId();
    // Get the language candidates in order and add the default language to it.
    $target_type = $this->getFieldSetting('target_type');
    $ancestor_langcodes = $language_manager->getFallbackCandidates([
      'langcode' => $langcode,
      'operation' => 'find_entity_reference_revisions',
      'data' => ['entity_type' => $target_type, 'entity_id' => $entity_id],
    ]);
    $default_langcode = $language_manager->getDefaultLanguage()->getId();
    if (!in_array($default_langcode, $ancestor_langcodes)) {
      $ancestor_langcodes = array_merge(
        $ancestor_langcodes,
        [$default_langcode => $default_langcode]
      );
    }
    $entity_storage = $this->entityTypeManager->getStorage($target_type);
    // This only makes sense for revisionable entities that support translation.
    if (!$entity_storage instanceof TranslatableRevisionableStorageInterface) {
      return $entity_storage instanceof RevisionableStorageInterface
        ? $entity_storage->getLatestRevisionId($entity_id)
        : $entity_id;
    }
    foreach ($ancestor_langcodes as $ancestor_langcode) {
      $revision_id = $entity_storage->getLatestTranslationAffectedRevisionId(
        $entity_id,
        $ancestor_langcode
      );
      if ($revision_id) {
        return $revision_id;
      }
    }
    return $entity_storage->getLatestRevisionId($entity_id);
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $build = parent::viewElements($items, $langcode);
    // If we are already previewing there is no need to show the indicator
    // either.
    if ($this->canSeeIndicator && !$this->isPreviewing) {
      $build = array_map(function (array $element) {
        $entity = $element['#' . $element['#theme']];
        $message = $this->t('Embedded entity has new draft');
        return $entity->isLatestTranslationAffectedRevision()
          ? $element
          : array_merge(
            $element,
            [
              'has_draft' => [
                '#type' => 'html_tag',
                '#tag' => 'span',
                '#attributes' => [
                  'class' => ['entity-reference-preview__has-draft'],
                  'title' => $message,
                ],
                '#attached' => ['library' => ['entity_reference_preview/indicator']],
                '#value' => $message,
              ],
            ]
          );
      }, $build);
    }
    $cache_contexts = ['user.permissions', 'entity_reference_preview'];
    (new CacheableMetadata())
      ->setCacheContexts($cache_contexts)
      ->applyTo($build);
    return $build;
  }

}
